// $Revision: 1.5 $
/*
    RJKDOC - A system for program documentation
    Copyright (C) 2000  Richard Kettlewell

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "rjkdoc.h"
#include "Document.h"

static const char __attribute__((unused)) cvs[] = "$Header: /cvs/rjkdoc/Document.cc,v 1.5 2003/06/08 22:28:14 richard Exp $";

// constructors and destructors ------------------------------------------------

Document::Document()
    : chapterType(chapter),
      depth(-1) {
}

Document::~Document() {
    assert(depth == -1);
}

// section structure -----------------------------------------------------------

// set the chapter type

void Document::setChapterType(ChapterType c) {
    if(chapterType != c)
	if(sections.size() != 0)
	    sections[0] = 0;
    chapterType = c;
}

// go down

wstring Document::down() {
    wstring name;			// the section name
    static wchar_t appnames[] = L".ABCDEFGHIJKLMNOPQRSTUVWXYZ";

    // go deeper
    depth++;

    // must have a parent
    assert(depth <= (int)sections.size());

    // extend the vector
    if(depth == (int)sections.size())
	sections.push_back(0);

    // increase section number
    sections[depth]++;

    // reset the subsection number
    if(depth + 1 < (int)sections.size())
	sections[depth + 1] = 0;

    // construct the section name
    for(int n = 0; n <= depth; n++) {
	if(n != 0) name += '.';
	if(n == 0 && chapterType == appendix) {
	    assert(sections[n] <= 26);
	    name += appnames[sections[n]];
	} else {
	    wchar_t buffer[64];

	    int_to_wstr(sections[n], buffer);
	    name += buffer;
	}
    }
    // chapter and appendix names have a final dot
    if(depth == 0) name += L'.';
    return name;
}

// go up

void Document::up() {
    depth--;
    assert(depth >= -1);
}
    
// handle events ---------------------------------------------------------------

// start of an element

void Document::startElement(const StartElementEvent &event) {
    wstring eventName = convertString(event.gi);
    
    if(debug)
	outputString(stderr, "stderr", L'<' + eventName + L">\n");
    elements.push_front(eventName);
    // at the start of a section, go down
    if(isSectionEvent(eventName)) {
	if(eventName == L"APPENDIX")
	    setChapterType(appendix);
	sectionNumber = down();
	// pick section names
	if(eventName == L"MANPAGE") {
	    wstring name, section;
	    
	    findAttributeValue(event, L"NAME", name);
	    findAttributeValue(event, L"SECTION", section);
	    sectionName = name + L'(' + section + L')';
	}
	
	if(eventName == L"NAME") sectionName = L"Name";
	if(eventName == L"SYNOPSIS") sectionName = L"Synopsis";
	if(eventName == L"DESCRIPTION") sectionName = L"Description";
    }

    // at the start of new inline content, clear the paragraph
    if(eventName == L"TITLE"
       || eventName == L"AUTHOR"
       || eventName == L"HEADING"
       || eventName == L"P"
       || eventName == L"PRE"
       || eventName == L"TAG")
	para.clear();
}

// data of various kinds

void Document::data(const DataEvent &event) {
    para.add(convertString(event.data));
}

void Document::sdata(const SdataEvent &event) {
    para.add(convertString(event.text));
}

// end element

void Document::endElement(const EndElementEvent &event) {
    wstring eventName = convertString(event.gi);
    
    if(debug)
	outputString(stderr, "stderr", L"</" + eventName + L">\n");
    if(isSectionEvent(eventName))
	up();
    if(eventName == L"HEADING")
	sectionName = para.getTidyText();
    elements.pop_front();
}

int Document::inside(const wstring &element) {
    for(list<wstring>::const_iterator p = elements.begin();
	p != elements.end();
	++p) {
	if(*p == element) return 1;
    }
    return 0;
}

// adding text -----------------------------------------------------------------

// add many lines

void Document::add(const list<wstring> &l, Part part) {
    for(list<wstring>::const_iterator p = l.begin();
	p != l.end();
	++p) {
	add(*p, part);
    }
}

void Document::addf(const list<wstring> &l, Part part) {
    for(list<wstring>::const_iterator p = l.begin();
	p != l.end();
	++p) {
	addf(*p, part);
    }
}

// utilities -------------------------------------------------------------------

// convert a CharString to a wstring

wstring Document::convertString(const CharString &s) {
    wstring t;
    
    for(size_t n = 0; n < s.len; n++) t += wchar_t(s.ptr[n]);
    return t;
}

// find an attribute value, return 1 on success or 0 on failure

int Document::findAttributeValue(const StartElementEvent &event,
				 const wstring &attributeName,
				 wstring &attributeValue) {
    wstring r;

    for(size_t n = 0; n < event.nAttributes; n++) {
	wstring name = convertString(event.attributes[n].name);

	if(name != attributeName) continue;
	switch(event.attributes[n].type) {
	case Attribute::cdata:
	    for(size_t m = 0; m < event.attributes[n].nCdataChunks; m++)
		r += convertString(event.attributes[n].cdataChunks[m].data);
	    attributeValue = r;
	    return 1;
	case Attribute::tokenized:
	    attributeValue = convertString(event.attributes[n].tokens);
	    return 1;
	case Attribute::implied:
	    return 0;
	default:
	    fprintf(stderr, "attribute type=%d\n",
		    (int)event.attributes[n].type);
	    abort();
	}
    }
    return 0;
}

// return 1 if the event name corresponds to a section, 0 otherwise

int Document::isSectionEvent(const wstring &event) {
    return (event == L"CHAPTER"
	    || event == L"APPENDIX"
	    || event == L"SECTION"
	    || event == L"MANPAGE"
	    || event == L"S1"
	    || event == L"S2"
	    || event == L"S3"
	    || event == L"NAME"
	    || event == L"SYNOPSIS"
	    || event == L"DESCRIPTION");
}

// find the document structure -------------------------------------------------

// XXX make the map sorted by section number
// (currently it is lexically sorted, which is wrong)

class ScanDocument : public Document {
private:
    map<wstring, wstring> &idMap;
    map<wstring, wstring> &sectionMap;
public:
    // the constructor remembers where to put the answers
    ScanDocument(map<wstring, wstring> &idMap_,
		 map<wstring, wstring> &sectionMap_)
	: idMap(idMap_),
	  sectionMap(sectionMap_) {
    }

    // handle start tags
    void startElement(const StartElementEvent &event) {
	Document::startElement(event);
	wstring eventName = convertString(event.gi);

	wstring id;
	if(findAttributeValue(event, L"ID", id))
	    idMap[id] = sectionNumber;

	if(eventName == L"MANPAGE"
	   || eventName == L"NAME"
	   || eventName == L"SYNOPSIS"
	   || eventName == L"DESCRIPTION")
	    sectionMap[sectionNumber] = sectionName;
    }

    // handle end tags
    void endElement(const EndElementEvent &event) {
	Document::endElement(event);
	wstring eventName = convertString(event.gi);

	if(eventName == L"HEADING")
	    sectionMap[sectionNumber] = sectionName;
    }
};

// construct a map of ID values to section numbers

void Document::findStructure(const char *path,
			     map<wstring, wstring> &idMap,
			     map<wstring, wstring> &sectionMap) {
    ScanDocument d(idMap, sectionMap);
    ParserEventGeneratorKit parserKit;

    EventGenerator *egp = parserKit.makeEventGenerator(1, (char **)&path);
    unsigned e = egp->run(d);
    if(e) {
	fprintf(stderr, "%u errors processing %s\n", e, path);
	exit(1);
    }
    delete egp;
}

// Local Variables:
// mode:c++
// comment-column:40
// End:
