// $Revision: 1.10 $
/*
    RJKDOC - A system for program documentation
    Copyright (C) 2000  Richard Kettlewell

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "rjkdoc.h"
#include "HTMLDocument.h"

static const char __attribute__((unused)) cvs[] = "$Header: /cvs/rjkdoc/HTMLDocument.cc,v 1.10 2003/06/20 21:28:34 richard Exp $";

// constructor and destructor --------------------------------------------------

// constructor

HTMLDocument::HTMLDocument(map<wstring, wstring> idMap_)
    : idMap(idMap_),
      manpage(0) {
    // set the doctype
    add(L"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">", titlepage);
    add(L"<!-- mechanically generated file - do not edit -->", titlepage);
    add(L"<html>", titlepage);
    rule(contents);
    add(L"<H2>Contents</H2>", contents);
    add(L"<UL>", contents);
    add(L"</UL>", body);
    rule(body);
    // set the default stylesheet
    style.push_back(L"<STYLE TYPE=\"text/css\">");
    style.push_back(L"  .FILE { font-style: italic }");
    style.push_back(L"  .PROG { font-weight: bold }");
    style.push_back(L"  .PROC { font-weight: bold }");
    style.push_back(L"  .OPTION { font-weight: bold }");
    style.push_back(L"  .LIT { font-family: monospace }");
    style.push_back(L"  .VAR { font-style: italic }");
    style.push_back(L"  .TITLE { text-align: center }");
    style.push_back(L"  .AUTHOR { text-align: center }");
    style.push_back(L"  .ABSTRACT { text-align: left; padding-left: 10%; padding-right: 10% }");
    style.push_back(L"  .COPYRIGHT { text-align: left }");
    style.push_back(L"  .MANPAGE { margin-left: 0.2in }");
    style.push_back(L"  EM { font-weight: bold; font-style: italic }");
    style.push_back(L"</STYLE>");
}

HTMLDocument::~HTMLDocument() {
}

void HTMLDocument::setStyle(const list<wstring> &newStyle) {
    style = newStyle;
}

// handle events ---------------------------------------------------------------

// start of element

void HTMLDocument::startElement(const StartElementEvent &event) {
    Document::startElement(event);
    wstring eventName = convertString(event.gi);

    if(eventName == L"MANPAGE") {
	manpage = 1;
	heading(sectionNumber + L' ' + sectionName);
    }
    
    if(eventName == L"DESCRIPTION"
       || eventName == L"SYNOPSIS"
       || eventName == L"NAME")
	heading(sectionName);

    if(eventName == L"COPYRIGHT")
	add(L"<DIV CLASS=COPYRIGHT>", titlepage);

    if(eventName == L"ABSTRACT")
	add(L"<DIV CLASS=ABSTRACT>", titlepage);

    if(eventName == L"TAGLIST")
	add(openTag(L"DL"));

    if(eventName == L"LIST")
	add(openTag(L"UL"));
    
    // FIXME: we want to know the deepest kind of list that we are in
    if(eventName == L"ITEM")
	if(inside(L"LIST")) add(L"<LI>");
	else add(L"<DD>");

    if(eventName == L"LIT")
	para.add(L"<SPAN CLASS=LIT>");

    if(eventName == L"VAR")
	para.add(L"<SPAN CLASS=VAR>");

    if(eventName == L"OPTION")
	para.add(L"<SPAN CLASS=OPTION>");

    if(eventName == L"FILE")
	para.add(L"<SPAN CLASS=FILE>");

    if(eventName == L"PROG")
	para.add(L"<SPAN CLASS=PROG>");
    if(eventName == L"PROC")
	para.add(L"<SPAN CLASS=PROC>");

    if(eventName == L"EM")
	para.add(L"<EM>");

    if(eventName == L"XREF") {
	wstring ref;
	map<wstring, wstring>::const_iterator p;
	
	findAttributeValue(event, L"REF", ref);
	p = idMap.find(ref);
	if(p == idMap.end()) {
	    fprintf(stderr, "WHOOPS!  Unmatched reference\n");
	    exit(1);
	}
	para.add(wstring(L"<A HREF=\"#")
		 + (*p).second + L"\">section " + (*p).second + L"</A>");
    }

    if(eventName == L"URL") {
	wstring name;
	
	findAttributeValue(event, L"NAME", name);
	para.add(L"<A HREF=\"" + name + L"\">" + name + L"</A>");
    }

    if(eventName == L"MANREF") {
	wstring name;
	wstring section;
	
	findAttributeValue(event, L"NAME", name);
	findAttributeValue(event, L"SECTION", section);
	para.add(L"<B>" + name + L"</B>(" + section + L")");
    }

    if(eventName == L"EMAIL")
	para.addf(L"<");

    if(eventName == L"OPTIONAL")
	para.add(L"[");
}

// end of element

void HTMLDocument::endElement(const EndElementEvent &event) {
    Document::endElement(event);
    wstring eventName = convertString(event.gi);

    if(eventName == L"MANPAGE")
      manpage = 0;
    
    if(eventName == L"TITLE") {
	add(L"<HEAD>", titlepage);
	add(wstring(L"<TITLE>")
	    + Paragraph::trim(para.getRawText())
	    + L"</TITLE>",
	    titlepage);
	add(style, titlepage);
	add(L"</HEAD>", titlepage);
	add(L"<BODY>", titlepage);
	add(wstring(L"<H1 CLASS=TITLE>")
	    + Paragraph::trim(para.getRawText())
	    + L"</H1>", titlepage);
    }
    if(eventName == L"AUTHOR") {
	add(wstring(L"<P CLASS=AUTHOR>")
	    + Paragraph::trim(para.getRawText())
	    + L"</P>", titlepage);
    }

    if(eventName == L"COPYRIGHT"
       || eventName == L"ABSTRACT")
	add(L"</DIV>", titlepage);
    
    if(eventName == L"HEADING") {
	if(manpage)
	    heading(sectionName);
	else
	    heading(sectionNumber + L' ' + sectionName);
    }

    if(eventName == L"CHAPTER"
       || eventName == L"APPENDIX") {
	rule();
	//blank(contents);
    }

    if(eventName == L"TAGLIST")
	add(L"</DL>");
 
    if(eventName == L"LIST")
	add(L"</UL>");
 
    if(eventName == L"TAG")
	add(wstring(L"<DT>") + Paragraph::trim(para.getRawText()) + L"</DT>");

    if(eventName == L"ITEM")
	if(inside(L"LIST")) add(L"</LI>");
	else add(L"</DD>");

    if(eventName == L"FILE"
       || eventName == L"LIT"
       || eventName == L"VAR"
       || eventName == L"PROG"
       || eventName == L"PROC"
       || eventName == L"OPTION")
	para.add(L"</SPAN>");

    if(eventName == L"EM")
	para.add(L"</EM>");
    
    if(eventName == L"P")
	add(openTag(L"P")
	    + Paragraph::trim(para.getRawText())
	    + L"</P>",
	    inside(L"TITLEPAGE") ? titlepage : body);
    
    if(eventName == L"PRE") {
	add(openTag(L"PRE") + para.getRawText() + L"</PRE>");
    }
    if(eventName == L"EMAIL")
	para.addf(L">");

    if(eventName == L"OPTIONAL")
	para.add(L"]");

    if(eventName == L"BODY")
	add(L"</BODY>");
    if(eventName == L"RJKDOC")
	add(L"</HTML>");
}

void HTMLDocument::data(const DataEvent &event) {
    para.addf(convertString(event.data));
}

void HTMLDocument::sdata(const SdataEvent &event) {
    para.addf(convertString(event.text));
}

// tools -----------------------------------------------------------------------

void HTMLDocument::heading(const wstring &s) {
    wchar_t level = '2' + sectionDepth();

    add(wstring(L"<H") + level + L">"
	+ L"<A NAME=\"" + sectionNumber + L"\" HREF=\"#" + sectionNumber + L"\">"
	+ s
	+ L"</A>"
	+ L"</H" + level + L">");
    // add to the contents
    if(sectionDepth() < 2)
	add(wstring(L"<LI>")
	    + L"<A HREF=\"#" + sectionNumber + L"\">"
	    + sectionNumber + L' ' + sectionName
	    + L"</A>"
	    + L"</LI>",
	    contents);
}

wstring HTMLDocument::openTag(const wchar_t *tag) {
    wstring s = L"<";

    s.append(tag);
    if(manpage)
	s.append(L" CLASS=MANPAGE");
    s.append(L">");
    return s;
}

// static methods --------------------------------------------------------------

// convert to HTML

void HTMLDocument::convert(const char *path, const char *output,
			   FILE *fp, const char *stylepath) {
    map<wstring, wstring> idMap_, sectionMap_;

    findStructure(path, idMap_, sectionMap_);
    
    HTMLDocument d(idMap_);
    if(stylepath) {
	FILE *in;
	char buffer[4096], *ptr;
	list<wstring> style_;

	if(!(in = fopen(stylepath, "r"))) {
	    perror(stylepath);
	    exit(1);
	}
	while(fgets(buffer, sizeof buffer, in)) {
	    if((ptr = strchr(buffer, '\n'))) {
		*ptr = 0;
	    }
	    style_.push_back(str_to_wstr(string(buffer)));
	}
	if(ferror(in)) {
	    perror(stylepath);
	    exit(1);
	}
	fclose(in);
	d.setStyle(style_);
    }
    ParserEventGeneratorKit parserKit;

    EventGenerator *egp = parserKit.makeEventGenerator(1, (char **)&path);
    unsigned e = egp->run(d);
    if(e) {
	fprintf(stderr, "%u errors processing %s\n", e, path);
	exit(1);
    }
    delete egp;
    for(int part = 0; part < PARTS; ++part)
	for(list<wstring>::const_iterator line = d.parts[part].begin();
	    line != d.parts[part].end();
	    ++line) {
	    outputString(fp, output, *line);
	    outputString(fp, output, L"\n");
	}
}

// Local Variables:
// mode:c++
// c-basic-offset:4
// comment-column:40
// End:
