// $Revision: 1.10 $
/*
    RJKDOC - A system for program documentation
    Copyright (C) 2000  Richard Kettlewell

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "rjkdoc.h"
#include "ManDocument.h"

static const char __attribute__((unused)) cvs[] = "$Header: /cvs/rjkdoc/ManDocument.cc,v 1.10 2003/06/08 22:28:14 richard Exp $";

// constructor and destructor --------------------------------------------------

ManDocument::ManDocument(map<wstring, wstring> idMap_,
			 const string &dir_,
			 const wstring &date_)
    : idMap(idMap_),
      dir(dir_),
      fp(NULL),
      date(date_) {
}

ManDocument::~ManDocument() {
}

// start/end of elements -------------------------------------------------------

void ManDocument::startElement(const StartElementEvent &event) {
    Document::startElement(event);
    wstring eventName = convertString(event.gi);

    if(eventName == L"MANPAGE") {
	wstring name;
	wstring section;
	string path;
	
	findAttributeValue(event, L"NAME", name);
	findAttributeValue(event, L"SECTION", section);
	path = string(dir) + '/' + wstr_to_str(name)
	    + '.' + wstr_to_str(section);
	if(!(output = strdup(path.c_str()))) {
	    perror("strdup");
	    exit(1);
	}
	if(!(fp = fopen(output, "w"))) {
	    perror(output);
	    exit(1);
	}
	wstring thline = L".TH ";
	thline += troffQuote(name);
	thline += L' ';
	thline += troffQuote(section);
	thline += L' ';
	thline += troffQuote(date);
	thline += L' ';
	thline += troffQuote(L"");	// source
	thline += L' ';
	thline += troffQuote(title);
	thline += L'\n';
	outputString(fp, output,
		     wstring(L".\\\" automatically generated file - do not edit\n")
	             + thline);		// XXX
	wasPara = 0;
    }
    if(fp) {
	if(eventName == L"NAME"
	   || eventName == L"SYNOPSIS"
	   || eventName == L"DESCRIPTION") {
	    outputString(fp, output, wstring(L".SH ") + eventName + L'\n');
	    wasPara = 0;
	} else if(eventName == L"MANREF") {
	    wstring name;
	    wstring section;
	    
	    findAttributeValue(event, L"NAME", name);
	    findAttributeValue(event, L"SECTION", section);
	    para.add(wstring(L"\\fB") + name + L"\\fR(" + section + L')');
	} else if(eventName == L"XREF") {
	    wstring ref;
	    
	    findAttributeValue(event, L"REF", ref);
	    para.add(ref);
	} else if(eventName == L"URL") {
	    wstring name;
	    
	    findAttributeValue(event, L"NAME", name);
	    para.add(wstring(L"<\\fB") + troffEscape(name) + L"\\fR>");
	} else if(eventName == L"FILE"
		  || eventName == L"VAR") {
	    openStyle(L'I');
	} else if(eventName == L"EM"
		  || eventName == L"PROC"
		  || eventName == L"PROG"
		  || eventName == L"OPTION") {
	    openStyle(L'B');
	} else if(eventName == L"LIT") {
	    openStyle(L'B');
	} else if(eventName == L"OPTIONAL") {
	    openStyle(L'R');
	    para.add(L'[');
	    closeStyle();
	} else if(eventName == L"ITEM" && inside(L"LIST")) {
	    outputString(fp, output, L".TP\n");
	    outputString(fp, output, L".B o\n");
	    wasPara = 0;
	}	    
    }
}

void ManDocument::endElement(const EndElementEvent &event) {
    Document::endElement(event);
    wstring eventName = convertString(event.gi);

    if(eventName == L"TITLE")
	title = para.getTidyText();
    if(eventName == L"MANPAGE") {
	if(fclose(fp) < 0) {
	    perror(output);
	    exit(1);
	}
	fp = NULL;
	free(output);
	output = NULL;
    }
    if(fp) {
	if(eventName == L"HEADING") {
	    if(sectionDepth() == 2) outputString(fp, output, L".SH ");
	    else outputString(fp, output, L".SS ");
	    wstring name = para.getTidyText();
	    for(wstring::size_type n = 0; n < name.length(); n++)
		name[n] = towupper(name[n]);
	    outputString(fp, output, troffQuote(name) + L'\n');
	    wasPara = 0;
	} else if(eventName == L"P") {
	    wstring text = para.getRawText();
	    wstring::size_type n = 0;
	    wstring ftext;
	    int offset = 0;
	    int lines = 0;
	    
	    while(n < text.length()) {
		if(iswspace(text[n])) {
		    if((offset == 0)
		       || n + 1 == text.length()
		       || (n + 1 < text.length()
			   && iswspace(text[n+1]))) {
			n++;
			continue;
		    }
		    if(offset > 64) {
			ftext += L'\n';
			lines++;
			offset = 0;
		    } else {
			ftext += L' ';
			offset++;
		    }
		} else {
		    ftext += text[n];
		    offset++;
		}
		n++;
	    }
	    if(offset != 0) {
		ftext += L'\n';
		lines++;
	    }
	    if(lines != 0) {
		if(wasPara)
		    if(inside(L"ITEM"))
			outputString(fp, output, L".IP\n");
		    else
			outputString(fp, output, L".PP\n");
		outputString(fp, output, ftext);
	    }
	    wasPara = 1;
	} else if(eventName == L"PRE") {
	    if(wasPara)
		outputString(fp, output, L".PP\n");
	    outputString(fp, output, L".nf\n");
	    outputString(fp, output, para.getRawText());
	    outputString(fp, output, L"\n.fi\n");
	    wasPara = 1;
	} else if(eventName == L"TAG") {
	    outputString(fp, output, L".TP\n");
	    outputString(fp, output, para.getTidyText() + L'\n');
	    wasPara = 0;
	} else if(eventName == L"FILE"
		  || eventName == L"EM"
		  || eventName == L"PROG"
		  || eventName == L"PROC"
		  || eventName == L"VAR"
		  || eventName == L"LIT"
		  || eventName == L"OPTION") {
	    closeStyle();
	} else if(eventName == L"OPTIONAL") {
	    openStyle(L'R');
	    para.add(L']');
	    closeStyle();
	}
    }
}

// data of various kinds

void ManDocument::data(const DataEvent &event) {
    para.add(troffEscape(convertString(event.data)));
}

void ManDocument::sdata(const SdataEvent &event) {
    para.add(troffEscape(convertString(event.text)));
}

// style management

wchar_t ManDocument::baseStyle() const {
    return L'R';
}

void ManDocument::setStyle(wchar_t s) {
    para.add(wstring(L"\\f") + s);
}

// troff support ---------------------------------------------------------------

wstring ManDocument::troffEscape(const wstring &s) {
    wstring escaped;

    for(wstring::size_type n = 0; n < s.length(); ++n) switch(s[n]) {
    default:
	escaped += s[n];
	break;
    case '.':
    case '\'':
	escaped += L"\\&";
	escaped += s[n];
	break;
    case L'\\':
	escaped += L"\\e";
	break;
    }
    return escaped;
}

wstring ManDocument::troffQuote(const wstring &s) {
    wstring quoted;

    // XXX spot strings we don't need to quote
    quoted += L'"';
    for(wstring::size_type n = 0; n < s.length(); ++n) switch(s[n]) {
    default:
	quoted += s[n];
	break;
    case L'"':
	quoted += L"\\\"";
	break;
    }
    quoted += L'"';
    return quoted;
    
}

// convert to man pages

void ManDocument::convert(const char *path, const char *dir_) {
    struct stat stat_buf;
    char tb[128];
    map<wstring, wstring> idMap_, sectionMap;

    findStructure(path, idMap_, sectionMap);
    
    if(stat(path, &stat_buf) < 0) {
	perror(path);
	exit(1);
    }
    strftime(tb, sizeof tb, "%Y-%m-%d", localtime(&stat_buf.st_mtime));
    ManDocument d(idMap_, dir_, str_to_wstr(tb));
    ParserEventGeneratorKit parserKit;

    EventGenerator *egp = parserKit.makeEventGenerator(1, (char **)&path);
    unsigned e = egp->run(d);
    if(e) {
	fprintf(stderr, "%u errors processing %s\n", e, path);
	exit(1);
    }
    delete egp;
}

// Local Variables:
// mode:c++
// comment-column:40
// End:
