// $Revision: 1.5 $
/*
    RJKDOC - A system for program documentation
    Copyright (C) 2000  Richard Kettlewell

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "rjkdoc.h"
#include "Paragraph.h"

static const char __attribute__((unused)) cvs[] = "$Header: /cvs/rjkdoc/Paragraph.cc,v 1.5 2000/01/21 14:17:21 richard Exp $";

Paragraph::Paragraph(const wstring &s)
    : raw(s) {
}

Paragraph::Paragraph() {
}

void Paragraph::add(const wstring &s) {
    for(wstring::size_type n = 0; n < s.length(); n++) add(s[n]);
}

void Paragraph::add(wchar_t c) {
    if(c == L'\r') c = L'\n';
    raw += c;
}

// format using justificaiton `j' in `width' characters across, with
// `prefix' spaces at the start of each line

list<wstring> Paragraph::format(Justification j, int width, int indent) const {
    wstring tidy = getTidyText();
    wstring::size_type len = tidy.length();
    list<wstring> r;

    wstring::size_type n = 0;
    while(n < len) {
	if(tidy[n] == L' ') {
	    n++;
	    continue;
	}
	wstring::size_type lineend = n + width;
	// n+width is the first character we won't print this line
	if(lineend < len) {
	    // partial line - find the final space
	    while(lineend > n && tidy[lineend] != L' ') --lineend;
	    if(lineend == n) {
		// There is a word longer than a line here.  We'll let
		// it overrun
		while(lineend < len && tidy[lineend] != L' ') ++lineend;
	    }
	    // now lineend either points at a space or at the end of
	    // the line.
	} else {
	    // this must be the last line
	    lineend = len;
	}
	wstring::size_type linelength = lineend - n;
	int spaces;
	switch(j) {
	case left:
	    spaces = indent;
	    break;
	case right:
	    spaces = indent + (width - linelength);
	    break;
	case centre:
	    spaces = indent + (width - linelength) / 2;
	    break;
	default:
	    abort();
	}
	wstring line;
	while(spaces-- > 0) line += L' ';
	line.append(tidy, n, linelength);
	r.push_back(line);
	n = lineend;
    }
    return r;
}

wstring Paragraph::getTidyText() const {
    wstring s;
    wstring::size_type n;
    int last_was_whitespace = 1;

    for(n = 0; n < raw.length(); n++) {
	if(iswspace(raw[n])) {
	    if(!last_was_whitespace) {
		s += L' ';
		last_was_whitespace = 1;
	    }
	} else {
	    s += raw[n];
	    last_was_whitespace = 0;
	}
    }
    if(last_was_whitespace && s.length() != 0) {
	s.erase(s.length() - 1, 1);
    }
    return s;
}

wstring Paragraph::getRawText() const {
    wstring::size_type e;

    for(e = raw.length(); e > 0 && iswspace(raw[e-1]); --e);
    return wstring(raw, 0, e);
}

void Paragraph::clear() {
    raw = L"";
}

wstring Paragraph::trim(const wstring &s) {
    wstring::size_type l = s.length(), n;
    wstring ret;
    int start_of_line = 1;
    
    n = 0;
    while(n < l) {
	if(!(start_of_line && iswspace(s[n]) && s[n] != L'\n')) {
	    start_of_line = (s[n] == L'\n');
	    ret += s[n];
	}
	n++;
    }
    return ret;
}


// Local Variables:
// mode:c++
// comment-column:40
// End:
