// $Revision: 1.7 $
/*
    RJKDOC - A system for program documentation
    Copyright (C) 2000  Richard Kettlewell

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "rjkdoc.h"
#include "TextDocument.h"

static const char __attribute__((unused)) cvs[] = "$Header: /cvs/rjkdoc/TextDocument.cc,v 1.7 2003/06/08 22:28:14 richard Exp $";

// constructor and destructor --------------------------------------------------

// constructor

TextDocument::TextDocument(map<wstring, wstring> idMap_,
			   int width_, int lm, int rm)
    : idMap(idMap_),
      width(width_ - rm - lm),
      indent(lm),
      bullet(0) {
    rule(contents);
    blank(contents);
    add(L"Contents", contents);
    blank(contents);
    rule(body);
    blank(body);
}

TextDocument::~TextDocument() {
}

// handle events ---------------------------------------------------------------

// start of element

void TextDocument::startElement(const StartElementEvent &event) {
    Document::startElement(event);
    wstring eventName = convertString(event.gi);

    if(eventName == L"MANPAGE"
       || eventName == L"DESCRIPTION"
       || eventName == L"SYNOPSIS"
       || eventName == L"NAME") {
	add(sectionNumber + L' ' + sectionName);
	blank();
	if(sectionDepth() < 2) {
	    add(sectionNumber + L' ' + sectionName, contents);
	}
    }
    if(eventName == L"TAG")
	indent += 2;
    if(eventName == L"ITEM") {
	if(inside(L"LIST")) bullet = L'o';
	indent += 4;
    }

    if(eventName == L"LIT")
	para.add(L'`');

    if(eventName == L"VAR")
	para.add(L'<');

    if(eventName == L"EMAIL")
	para.add(L'<');

    if(eventName == L"OPTIONAL")
	para.add(L'[');

    if(eventName == L"OPTION")
	openStyle(L'B');

    if(eventName == L"EM")
	openStyle(L'B');

    if(eventName == L"FILE")
	openStyle(L'U');

    if(eventName == L"PROC")
	openStyle(L'B');

    if(eventName == L"PROG")
	openStyle(L'B');
    
    if(eventName == L"XREF") {
	wstring ref;
	map<wstring, wstring>::const_iterator p;
	
	findAttributeValue(event, L"REF", ref);
	p = idMap.find(ref);
	if(p == idMap.end()) {
	    fprintf(stderr, "WHOOPS!  Unmatched reference\n");
	    exit(1);
	}
	para.add(L"section " + escapeString((*p).second));
    }

    if(eventName == L"URL") {
	wstring name;
	
	findAttributeValue(event, L"NAME", name);
	para.add(L'<' + escapeString(name) + L'>');
    }

    if(eventName == L"MANREF") {
	wstring name;
	wstring section;
	
	findAttributeValue(event, L"NAME", name);
	findAttributeValue(event, L"SECTION", section);
	openStyle(L'B');
	para.add(escapeString(name));
	closeStyle();
	para.add(escapeString(wstring(L"(") + section + L')'));
    }
}

// end of element

void TextDocument::endElement(const EndElementEvent &event) {
    Document::endElement(event);
    wstring eventName = convertString(event.gi);

    if(eventName == L"TITLE"
	|| eventName == L"AUTHOR")
	format(1, Paragraph::centre, titlepage);
    
    if(eventName == L"HEADING") {
	add(sectionNumber + L' ' + sectionName);
	blank();
	// add to the contents
	if(sectionDepth() < 2) {
	    add(sectionNumber + L' ' + sectionName, contents);
	}
    }

    if(eventName == L"CHAPTER"
       || eventName == L"APPENDIX") {
	rule();
	blank();
	blank(contents);
    }
    
    if(eventName == L"TAG")
	format(0);
    if(eventName == L"P")
	if(inside(L"TITLEPAGE"))
	    format(1, Paragraph::left, titlepage);
	else
	    format(1, Paragraph::left, body);

    if(eventName == L"PRE") {
	list<wstring> text = splitString(para.getRawText(), L'\n');
	if(text.size() != 0) {
	    for(list<wstring>::const_iterator p = text.begin();
		p != text.end();
		++p) {
		add(wstring(indent, L' ') + *p);
	    }
	    blank();
	}
    }

    if(eventName == L"TAG")
	indent -= 2;
    if(eventName == L"ITEM") {
	indent -= 4;
	bullet = 0;
    }

    if(eventName == L"LIT")
	para.add(L'\'');

    if(eventName == L"VAR")
	para.add(L'>');

    if(eventName == L"EMAIL")
	para.add(L'>');

    if(eventName == L"OPTIONAL")
	para.add(L']');

    if(eventName == L"OPTION"
       || eventName == L"FILE"
       || eventName == L"PROG"
       || eventName == L"PROC"
       || eventName == L"EM")
	closeStyle();
}

void TextDocument::data(const DataEvent &event) {
    para.add(escapeString(convertString(event.data)));
}

void TextDocument::sdata(const SdataEvent &event) {
    para.add(escapeString(convertString(event.text)));
}

// style management

wchar_t TextDocument::baseStyle() const {
    return L'R';
}

void TextDocument::setStyle(wchar_t s) {
    para.add(wstring(L"!") + s);
}

wstring TextDocument::escapeString(const wstring &s) {
    wstring ret;
    
    for(wstring::size_type n = 0; n < s.length(); ++n) {
	if(s[n] == L'!') ret += L'!';
	ret += s[n];
    }
    return ret;
}

// formatting text -------------------------------------------------------------

void TextDocument::format(int needsep,
			  Paragraph::Justification j,
			  Part part) {
    // format the text
    list<wstring> formatted = para.format(j, width - indent, indent);

    // add the text plus a blank line
    if(formatted.size()) {
	for(list<wstring>::iterator p = formatted.begin();
	    p != formatted.end();
	    ++p) {
	    if(bullet) {
		(*p)[indent - 2] = bullet;
		bullet = 0;
	    }
	    add(*p, part);
	}
	if(needsep) blank(part);
    }
}

// construct a map of ID values to section numbers

void TextDocument::convert(const char *path, const char *output, FILE *fp) {
    map<wstring, wstring> idMap_, sectionMap;

    findStructure(path, idMap_, sectionMap);
    
    TextDocument d(idMap_);
    ParserEventGeneratorKit parserKit;

    EventGenerator *egp = parserKit.makeEventGenerator(1, (char **)&path);
    unsigned e = egp->run(d);
    if(e) {
	fprintf(stderr, "%u errors processing %s\n", e, path);
	exit(1);
    }
    delete egp;
    for(int part = 0; part < PARTS; ++part)
	for(list<wstring>::const_iterator line = d.parts[part].begin();
	    line != d.parts[part].end();
	    ++line) {
	    outputString(fp, output, *line);
	    outputString(fp, output, L"\n");
	}
}

// Local Variables:
// mode:c++
// comment-column:40
// End:
